/*
 * XPathProcessor - ESIAL PROJET SD 2011
 * Copyright (C) 2011 Julien VAUBOURG / Tristan STEF
 * Contact <julien@vaubourg.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package Processeur;
import java.util.HashMap;
import java.util.regex.*;

public class AnalyseurLexical {
	private StringBuilder terminalBuffer;
	private TypeTerminal typeTerminal;
	private HashMap<String, String> attributsTrouves;
	
	public static enum TypeTerminal {
		B_OUVRANTE,
		B_FERMANTE,
		TEXTE,
		NT
	};

	public AnalyseurLexical() {
		reset();
	}

	/** 
	 * Tente de reconnaitre les terminaux de la grammaire au fur et a mesure que de nouveaux
	 * caracteres lui sont donnes. 
	 * @param c Nouveau caractere pour le buffer
	 * @return Un terminal (balise ouvrante/fermante, ou du texte brut) reconnu, ou null
	 * @throws ExceptionLexicale Lorsqu'une balise ne respecte pas la syntaxe XML ou ne fait pas parti de la grammaire 
	 */
	public String interpreter(char c) throws ExceptionLexicale {
		
		// Par defaut, c'est un Non-Terminal qui est renvoye
		String retour = null;
		typeTerminal = TypeTerminal.NT;
		attributsTrouves = new HashMap<String, String>();
 
		// Si un debut de balise est detecte, et que le buffer n'est pas vide,
		// il s'agit donc de texte brut
		if(c == '<' && !terminalBuffer.toString().matches("^\\s*$")) {
			retour = terminalBuffer.toString();
			typeTerminal = TypeTerminal.TEXTE;
			reset();
		}

		// Ajout du caractere au buffer
		terminalBuffer.append(c);
		
		// Si une fin de balise est detectee, on verifie si la balise globale est conforme
		if(c == '>') {
			
			// retour contient "balise" ou "/balise", ou null si elle n'est pas conforme
			if((retour = extraitBalise()) != null) {
				
				// Balise fermante, passage de "/balise" a "balise" 
				if(retour.charAt(0) == '/') {
					retour = retour.substring(1);
					typeTerminal = TypeTerminal.B_FERMANTE;
					
				// Balise ouvrante
				} else
					typeTerminal = TypeTerminal.B_OUVRANTE;
				
				// Teste si le terminal ainsi trouve fait partie de la grammaire
				try {
					UniteLexicale terminal = (UniteLexicale) Class.forName("Grammaire." + EntreesSorties.ajouterMajuscule(retour)).newInstance();
					
					for(String attribut : attributsTrouves.keySet())
						if(!terminal.peutAccepterAttribut(attribut)) {
							reset();

							throw new ExceptionLexicale("L'attribut " + attribut + " n'est pas reconnu par la grammaire comme un attribut valide de <" + retour + ">.");
						}
					
				} catch (InstantiationException e) {}
				  catch (IllegalAccessException e) {}
				  catch (ClassNotFoundException e) {
					reset();

					throw new ExceptionLexicale("<" + retour + "> est inconnu de la grammaire.");
				}

				// Reinitialisation du buffer chaque fois qu'un terminal est renvoye
				terminalBuffer = new StringBuilder();

			} else {
				Matcher trim = Pattern.compile("\\s*(.*)\\s*").matcher(terminalBuffer.toString());
				trim.find();
				reset();

				throw new ExceptionLexicale("[ " + trim.group(1) + " ] ne semble pas être une balise cohérente.");
			}
		}
		
		return retour;
	}
	
	/**
	 * Renvoit le type du terminal dernierement renvoye par le buffer via la fonction
	 * interpreter, ou NT si le dernier caractere n'a pas donne lieu a une reconnaissance de
	 * terminal.
	 * @return Type du terminal dernierement renvoye, ou NT
	 */
	public TypeTerminal getTypeTerminal() {
		return typeTerminal;
	}
	
	/**
	 * Renvoit les attributs trouves pour le dernier terminal renvoye par le buffer. 
	 * @return Derniers attributs trouves
	 */
	public HashMap<String, String> getAttributs() {
		return attributsTrouves;
	}

	/**
	 * Remet a zero la recherche et le buffer.
	 */
	private void reset() {
		terminalBuffer = new StringBuilder();
		attributsTrouves = new HashMap<String, String>(); 
	}
	
	/**
	 * Passe le buffer de "&lt;balise&gt;" a "balise" et de "&lt;/balise&gt; a /balise, ou
	 * renvoit null si aucune balise conforme n'a ete detectee.
	 * @return La balise sans ses chevrons, ou null
	 */
	private String extraitBalise() throws ExceptionLexicale {
		String patternArguments = "";
		Matcher chercherNbArguments, chercher;
		String balise = null;
		
		for(
			chercherNbArguments = Pattern.compile("=").matcher(terminalBuffer);
			chercherNbArguments.find();
			patternArguments += "\\s*([^=]+)=\"([^\"]+)\""
		);

		chercher = Pattern.compile("\\s*<(/?[\\w-]+)" + patternArguments + "\\s*>\\s*").matcher(terminalBuffer);
		
		if(chercher.find()) {
			balise = chercher.group(1);
			
			for(int i = 2; i <= chercher.groupCount(); i += 2) {
				if(attributsTrouves.containsKey(chercher.group(i))) {
					reset();

					throw new ExceptionLexicale("<" + balise + "> posséde plus d'un attribut " + chercher.group(i) + ".");
				}

				attributsTrouves.put(chercher.group(i), chercher.group(i+1));
			}
		}

		return balise;
	}
}