/*
 * XPathProcessor - ESIAL PROJET SD 2011
 * Copyright (C) 2011 Julien VAUBOURG / Tristan STEF
 * Contact <julien@vaubourg.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package Processeur;

import java.util.HashMap;
import Grammaire.Text;

public class AnalyseurSyntaxique {
	private UniteLexicale pointeur;

	/**
	 * Constructeur
	 * @param axiome Racine imposee du document (premier terminal a trouver)
	 */
	public AnalyseurSyntaxique(UniteLexicale axiome) {
		pointeur = axiome;
	}

	/**
	 * Ajoute un texte brut dans un terminal Texte, aux fils du terminal actuellement pointe
	 * par pointeur.
	 * @param texte Texte brut a ajouter en fils
	 * @throws ExceptionSyntaxique Lorsque le terminal actuellement pointe ne peut pas contenir de texte
	 */
	public void insererTerminalTexte(String texte) throws ExceptionSyntaxique {
		if(pointeur.estPremier("texte")) {
			UniteLexicale fils = new Text(texte);
			fils.setPere(pointeur);
			pointeur.inserer(fils);	
		} else
			throw new ExceptionSyntaxique(pointeur + " ne peut pas contenir de texte.");
	}

	/**
	 * Ajoute un nouveau terminal aux fils du terminal actuellement pointe par pointeur.
	 * @param terminal Nom du terminal a ajouter
	 * @param attributs Liste des attributs values
	 * @throws ExceptionSyntaxique Si la racine est mal definie, une balise fermante est manquante, ou l'ordre
	 * des balises est mauvais
	 * @throws InstantiationException Un noeud &lt;text&gt; a ete cree explicitement
	 */
	public void insererTerminal(String terminal, HashMap<String, String> attributs) throws ExceptionSyntaxique, InstantiationException {
		try {
			// Un noeud <text>...</text> ou <void>...</void> a ete trouve dans le fichier
			if(terminal.equals("text") || terminal.equals("void"))
				throw new InstantiationException("Erreur interne : Noeud <" + terminal + "> explicite interdit.");

			// Cas du pointeur qui pointe vers l'axiome
			if(pointeur.getPere() == null) {
				
				// Si le premier terminal trouve ne correspond pas a l'axiome, la racine est mal definie
				if(!pointeur.getNom().equals(terminal))
					throw new ExceptionSyntaxique("La racine du document doit être " + pointeur + " (et non <" + terminal + ">)");
				
				// Sinon, on ne fait rien, et on neutralise le pere de l'axiome pour indiquer qu'il a deja
				// ete consomme
				else
					pointeur.setPere(pointeur);
			
			// Si le pointeur accepte d'inserer ce terminal (il fait partie des suivants de son dernier, ou fait partie de ses premiers si
			// il n'a pas encore de fils)
			} else if(pointeur.peutInserer(terminal)) {
				
				// Instanciation du terminal avec la classe correspondante dans la grammaire
				UniteLexicale fils = (UniteLexicale) Class.forName("Grammaire." + EntreesSorties.ajouterMajuscule(terminal)).newInstance();

				fils.setPere(pointeur);
				fils.setAttributs(attributs);

				pointeur.inserer(fils);
				pointeur = fils;

			// Le terminal est refuse par le pointeur
			} else
				throw new ExceptionSyntaxique(pointeur + " ne peut accepter un <" + terminal + ">" + (pointeur.estSansFils() ? " en premier fils" : ", après un " + pointeur.dernier()));
		
		// Ne devrait jamais arriver : c'est une erreur lexicale (terminal inconnu de la grammaire)
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
			
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Tente de fermer le terminal du pointeur, en indiquant son nom.
	 * @param terminal Nom du terminal a fermer
	 * @throws ExceptionSyntaxique Si le terminal qu'on demande a fermer n'est pas celui qui est ouvert dans le pointeur
	 */
	public void fermer(String terminal) throws ExceptionSyntaxique {
		
		// Le terminal a fermer correspond a celui qui est actuellement ouvert
		if(pointeur.getNom().equals(terminal)) {
			pointeur.fermer();
			pointeur = pointeur.getPere();
			
		// ... ou pas. Un oubli ?
		} else
			throw new ExceptionSyntaxique("</" + pointeur.getPere().dernier().getNom() + "> manquant.");
	}
}