/*
 * XPathProcessor - ESIAL PROJET SD 2011
 * Copyright (C) 2011 Julien VAUBOURG / Tristan STEF
 * Contact <julien@vaubourg.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package Processeur;

import java.io.IOException;
import java.util.ArrayList;
import java.util.regex.*;
import Grammaire.Text;

public class InterpreteurXPath {

	public static void demarrer(UniteLexicale axiome, String fichierSortie) {
		String requeteXPath;
		
		try {
			// Boucle d'execution de requetes XPath donnees par l'utilisateur
			do {
				// Saisie de la requete
				requeteXPath = EntreesSorties.saisirRequeteXPath();
	
				if(!requeteXPath.isEmpty()) {
	
					// EXCEPTION : La requete est syntaxiquement malformee (exception xpath)
					ArrayList<UniteLexicale> noeuds;
	
					try {
	
						// Interpretation de la requete
						noeuds = InterpreteurXPath.interpreter(axiome, requeteXPath);
		
						// Aucun resultat
						if(noeuds.isEmpty())
							System.out.println("\n\t/!\\ Aucun résultat.\n");
						
						// Un ou plusieurs resultats
						else
							// Ecriture sur la sortie standard
							if(fichierSortie.isEmpty())
								for(UniteLexicale noeud : noeuds)
									System.out.print(noeud.xml());
	
							// Ecriture dans un fichier
							else {
								EntreesSorties.ecrire(fichierSortie, noeuds);
	
								System.out.println("\n(info) Résultats imprimés dans " + fichierSortie + ".");
								requeteXPath = "";
							}
	
					} catch (ExceptionXPath e) {
						System.out.println("\n\t/!\\ " + e.getMessage() + "\n");
					}
				}
	
			} while(!requeteXPath.isEmpty());
			
		} catch(IOException e) {
			System.err.println("Problème à l'écriture du fichier " + fichierSortie + ".");
		}
	}
	
	/**
	 * Cherche a repondre a la requete XPath a partir de l'axiome
	 * @param axiome Arbre complet (root)
	 * @param requete Requete XPath
	 * @return Liste des noeuds correspondants 
	 * @throws ExceptionXPath Requete XPath malformee
	 */
	public static ArrayList<UniteLexicale> interpreter(UniteLexicale axiome, String requete) throws ExceptionXPath {
		ArrayList<UniteLexicale> noeuds = new ArrayList<UniteLexicale>();
		noeuds.add(axiome);

		if(requete.matches("//?" + axiome.getNom()))
			return noeuds; 

		// Si la requete commence par preciser l'axiome, on retire cette partie,
		// sinon c'est juste l'occasion de retirer les blancs autour de la requete
		Matcher trim = Pattern.compile("^\\s*(//?([^/]+)(.*))\\s*$").matcher(requete);
		
		// On a bien trouve un semblant de requete
		if(trim.find()) {
			
			// Il s'agit d'un chemin absolu (ou pas) avec l'axiome en premier : on le vire
			if(trim.group(2).equals(axiome.getNom()))
				return interpreter(noeuds, trim.group(3));
			
			// Sinon, ce doit absolument etre un chemin relatif 
			else if(requete.startsWith("//"))
				return interpreter(noeuds, trim.group(1));
	
			// Sinon, c'est une erreur
			else
				return new ArrayList<UniteLexicale>();
		} else
			throw new ExceptionXPath(requete);	
	}
	
	/**
	 * Cherche recursivement a repondre a la requete XPath
	 * @param noeuds Liste de noeuds a partir desquels la requete doit etre interpretee
	 * @param requete Requete XPath
	 * @return Liste des noeuds correspondants
	 * @throws ExceptionXPath Requete XPath malformee
	 */
	private static ArrayList<UniteLexicale> interpreter(ArrayList<UniteLexicale> noeuds, String requete) throws ExceptionXPath {
		String requeteSuivante, requeteCourante, predicat;
		ArrayList<UniteLexicale> resultats = new ArrayList<UniteLexicale>();

		// Separation du bout de requete a tester (dans //personne/telephone, ce sera //personne) ainsi que
		// de l'eventuel predicat, et la fin de la requete qu'il faudra relancer sur les resultats obtenus
		Matcher chercher = Pattern.compile("^//?(@?\\w+)(?:\\(\\)|\\[(@?[\\w=\"']+)\\])?(.*)").matcher(requete);

		// Pattern XPath reconnu ?
		if(!chercher.find())
			throw new ExceptionXPath(requete);

		// Parties de la requete
		requeteCourante = chercher.group(1);
		predicat = chercher.group(2);
		requeteSuivante = chercher.group(3);
		
		// Requete elliptique : on descend parmi les descendants des noeuds a tester, jusqu'a ce qu'on
		// trouve un descendant du nom de ce qui est cherche
		if(requete.startsWith("//")) {
			for(UniteLexicale fils : noeuds)
				for(UniteLexicale filsIndirect : fils.getFilsIndirects())					
					if(filsIndirect.getNom().equalsIgnoreCase(requeteCourante))
						resultats.add(filsIndirect);
		
		// On demande la valeur d'un attribut (/@attribut)
		} else if(requete.startsWith("/@")) {
			String attribut = requete.substring(2);

			for(UniteLexicale fils : noeuds)
				if(fils.possedeAttribut(attribut))
					resultats.add(new Text(fils.getAttribut(attribut)));

		// Passage d'un noeud a l'autre : on recherche dans les fils directs des noeuds, ceux qui
		// correspondent a celui-ci
		} else if(requete.startsWith("/")) {
			for(UniteLexicale fils : noeuds)
				for(UniteLexicale filsDirect : fils.getFils())
					if(filsDirect.getNom().equalsIgnoreCase(requeteCourante))
						resultats.add(filsDirect);
		}
		
		// Filtrage avec l'eventuel predicat : les resultats obtenus ci-dessus vont etre epures
		// pour respecter les contraintes du predicat
		if(predicat != null) {
			
			// personne[3] : On veut la troisieme personne trouvee
			if(predicat.matches("^[0-9]+$")) {
				ArrayList<UniteLexicale> filtre = new ArrayList<UniteLexicale>();
				
				try {
					filtre.add(resultats.get(Integer.parseInt(predicat) - 1));
					resultats = filtre;

				// Index du predicat inexistant
				} catch(IndexOutOfBoundsException e) {
					resultats = new ArrayList<UniteLexicale>();
				}
			
			// telephone[@type... : predicats en fonction d'un attribut
			} else if(predicat.startsWith("@")) {

				// telephone[@type='gsm'] : on veut les telephones qui on gsm comme valeur de leur attribut type
				if(predicat.indexOf('=') > -1) {
					Matcher comparaison = Pattern.compile("^@(\\w+)=[\"'](\\w+)[\"']$").matcher(predicat);

					// Verification de la syntaxe exacte du predicat
					if(comparaison.find()) {
						String attribut = comparaison.group(1);
						String valeur = comparaison.group(2);

						for(int i = resultats.size() - 1; i >= 0; i--)
							if(!resultats.get(i).attributEquals(attribut, valeur))
								resultats.remove(i);
					} else
						throw new ExceptionXPath(predicat);
					
				// telephone[@type] : on veut les telephones qui ont un attribut type
				} else
					for(int i = resultats.size() - 1; i >= 0; i--)
						if(!resultats.get(i).possedeAttribut(predicat.substring(1)))
							resultats.remove(i);
	
			// personne[telephone] : On veut les personnes qui ont un telephone defini en fils direct 
			} else if(!predicat.isEmpty()) {
	
				for(int i = resultats.size() - 1; i >= 0; i--)
					if(!resultats.get(i).possedeFils(predicat))
						resultats.remove(i);
			}
		}
		
		// Fin de la recursion lorsque la requete initiale est totalement consommee
		return requeteSuivante.isEmpty() || resultats.isEmpty() ? resultats : interpreter(resultats, requeteSuivante);
	}
}
