/*
 * XPathProcessor - ESIAL PROJET SD 2011
 * Copyright (C) 2011 Julien VAUBOURG / Tristan STEF
 * Contact <julien@vaubourg.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package Processeur;

import java.util.*;
import java.util.Map.Entry;

public abstract class UniteLexicale {
	private ArrayList<UniteLexicale> fils;
	private ArrayList<String> premiers, suivants;
	private HashMap<String, String> attributs;
	private UniteLexicale pere = null;

	/**
	 * Constructeur (appele implicitement par java par les classes filles)
	 */
	public UniteLexicale() {
		fils = new ArrayList<UniteLexicale>();
		premiers = new ArrayList<String>();
		suivants = new ArrayList<String>();
		attributs = new HashMap<String, String>();
	}
	
	/**
	 * Defini le pere du terminal.
	 * @param pere Pere du terminal
	 */
	public void setPere(UniteLexicale pere) {
		this.pere = pere;
	}

	/**
	 * Renvoi le pere du terminal
	 * @return Pere du terminal
	 */
	public UniteLexicale getPere() {
		return pere;
	}

	/**
	 * Sert a indiquer tous les terminaux qui peuvent etre les premiers fils du terminal
	 * @param premiers Nom des terminaux
	 */
	protected void setPremiers(String...premiers) {
		for(String premier : premiers)
			this.premiers.add(premier);
	}

	/**
	 * Sert a indiquer tous les terminaux qui peuvent succeder a ce terminal
	 * @param suivants Nom des terminaux
	 */
	protected void setSuivants(String...suivants) {
		for(String suivant : suivants)
			this.suivants.add(suivant);
	}

	/**
	 * Sert a indiquer tous les attributs qui peuvent convenir a ce terminal
	 * @param suivants Nom des attributs (CDATA, facultatifs)
	 */
	protected void setAttributs(String...attributs) {
		for(String attribut : attributs)
			this.attributs.put(attribut, null);
	}
	
	/**
	 * Sert a renseigner tous les attributs du terminal (l'analyseur lexical aura pris soin
	 * de verifier qu'ils sont bien acceptables).
	 * @param attributs Attributs values
	 */
	protected void setAttributs(HashMap<String, String> attributs) {
		if(attributs != null)
			this.attributs = attributs;
	}

	/**
	 * Indique en fonction de ses premiers et des suivants de ses fils, si le terminal peut
	 * ajouter le nouveau terminal passe en parametre a ses fils.
	 * @param terminal Fils a ajouter a la suite des autres
	 * @return Respecte les premiers et suivants (peut etre ajoute ou non)
	 */
	public boolean peutInserer(String terminal) {
		return fils.isEmpty() && premiers.contains(terminal) ||
			!fils.isEmpty() && dernier().estSuivant(terminal);
	}

	/**
	 * Indique si ce terminal accepte cet attribut.
	 * @param attribut Attribut a tester
	 * @return Accepte ou non
	 */
	public boolean peutAccepterAttribut(String attribut) {
		return attributs.containsKey(attribut);
	}

	/**
	 * Fermeture de la balise 
	 * @throws ExceptionSyntaxique Des terminaux sont manquants dans la balise
	 */
	public void fermer() throws ExceptionSyntaxique {
		if(fils.isEmpty() && !estPremier("void") || !fils.isEmpty() && !dernier().estSuivant("void"))
			throw new ExceptionSyntaxique("Des terminaux sont manquants dans " + this);
	}

	/**
	 * Dernier fils ajoute
	 * @return Derniers fils ajoute
	 */
	public UniteLexicale dernier()  {
		return fils.get(fils.size() - 1);
	}
	
	/**
	 * Ajoute un fils (terminal) a la suite des autres
	 * @param terminal Fils a ajouter
	 */
	public void inserer(UniteLexicale terminal) {
		fils.add(terminal);
	}

	/**
	 * Permet de savoir un terminal donne est premier de celui-ci
	 * @param terminal Nom du terminal a tester
	 * @return Le terminal est premier ou non
	 */
	public boolean estPremier(String terminal) {
		return premiers.contains(terminal);
	}

	/**
	 * Permet de savoir un terminal donne est suivant de celui-ci
	 * @param terminal Nom du terminal a tester
	 * @return Le terminal est suivant ou non
	 */
	public boolean estSuivant(String terminal) {
		return suivants.contains(terminal);
	}
	
	/**
	 * Permet de savoir si la balise a deja des fils ou non.
	 * @return La balise a deja des fils ou non
	 */
	public boolean estSansFils() {
		return fils.isEmpty();
	}
	
	/**
	 * Nom de la balise representee par ce terminal, en minuscule.
	 * @return Nom de la balise
	 */
	public String getNom() {
		return getClass().getCanonicalName().substring(getClass().getCanonicalName().lastIndexOf('.') + 1).toLowerCase();
	}
	
	/**
	 * Permet de savoir si la balise contient du texte ou non.
	 * @return La balise contient du texte ou non
	 */
	public boolean estTerminalTexte() {
		return !fils.isEmpty() && dernier().estTexte();
	}

	/**
	 * Permet de savoir si la balise est du texte ou non.
	 * @return La balise est du texte ou non
	 */
	public boolean estTexte() {
		return getNom().equals("text");
	}
	
	/**
	 * &lt;balise&gt;...&lt;/balise&gt;
	 */
	public String toString() {
		return "<" + getNom() + ">...</" + getNom() + ">";
	}

	/**
	 * Renvoit les fils du noeud courant.
	 * @return Fils
	 */
	public ArrayList<UniteLexicale> getFils() {
		return fils;
	}
	
	/**
	 * Renvoit descendants du noeud courant
	 * @return Descendants
	 */
	public ArrayList<UniteLexicale> getFilsIndirects() {
		ArrayList<UniteLexicale> indirects = new ArrayList<UniteLexicale>();
		
		// Demande les fils des fils
		for(UniteLexicale unFils : fils) {
			indirects.add(unFils);
			indirects.addAll(unFils.getFilsIndirects());
		}
		
		return indirects;
	}
	
	/**
	 * Renvoit descendants du noeud courant du nom donne en parametre
	 * @return Descendants
	 */
	public ArrayList<UniteLexicale> getFilsIndirectsParNom(String nom) {
		ArrayList<UniteLexicale> indirects = new ArrayList<UniteLexicale>();

		// Demande les fils des fils
		for(UniteLexicale unFils : getFilsIndirects())
			if(unFils.getNom().equals(nom))
				indirects.add(unFils);

		return indirects;
	}
	
	/**
	 * Teste si le noeud contient le fils
	 * @param nom Nom du fils a tester
	 * @return Contient le fils ou non
	 */
	public boolean possedeFils(String nom) {
		for(UniteLexicale unFils : fils)
			if(unFils.getNom().equals(nom))
				return true;
		
		return false;
	}
	
	/**
	 * Teste si le noeud possede l'attribut
	 * @param attribut Attribut a tester
	 * @return L'attribut est acceptable ou non
	 */
	public boolean possedeAttribut(String attribut) {
		return attributs.containsKey(attribut);
	}
	
	/**
	 * Teste la valeur de l'attribut 
	 * @param attribut Attribut concerne
	 * @param valeur Valeur a tester 
	 * @return L'attribut a cette valeur ou non
	 */
	public boolean attributEquals(String attribut, String valeur) {
		return attributs.containsKey(attribut) && attributs.get(attribut).equals(valeur); 
	}

	/**
	 * Renvoit la valeur de l'attribut
	 * @param attribut Attribut concerne
	 * @return Valeur de l'attribut
	 */
	public String getAttribut(String attribut) {
		return attributs.get(attribut);
	}

	/**
	 * Sortie texte
	 * @return Arbre xml texte
	 */
	public String xml() {
		return xml(0);
	}
	
	/**
	 * Sortie texte recurcive de l'arbre du document XML stocke dans la SD.
	 * @param profondeur Profondeur dans l'arbre XML
	 * @return Arbre xml texte
	 */
	public String xml(int profondeur) {
		String retour, attributs = "", retrait = "";

		// Ecriture des attributs de la balise
		for(Entry<String, String> attribut : this.attributs.entrySet())
			attributs += " " + attribut.getKey() + "=\"" + attribut.getValue() + "\"";
		
		retour = "<" + getNom() + attributs + ">";

		// Indentation
		for(int i = 0; i < profondeur; retrait += "\t", i++);
		retour += estTerminalTexte() ? "" : "\n"; 

		// Fils (sans retour a la ligne pour les textes)
		for(UniteLexicale unFils : fils) {
			String xml = unFils.xml(profondeur + 1);
			retour += unFils.estTexte() ? xml.substring(0, xml.length() - 1) : xml; 
		}

		return retrait + retour + (estTerminalTexte() ? "" : retrait) + "</" + getNom() + ">\n";	
	}
}